#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#ifndef NO_UNISTD_H
#include <unistd.h>
#endif
#include <syslog.h>
#include <fcntl.h>
#include "wn.h"
#include "globdec.h"
#include "err.h"
#include "content.h"

#if STANDALONE
#define PRINT_ERROR( x, y)	fprintf( stderr, x, y);
#define PRINT1_ERROR( x)	fprintf( stderr, x);
#else
#define PRINT_ERROR( x, y)	printf( x, y);
#define PRINT1_ERROR( x)	printf( x);
#endif

extern char	*optarg;
extern int	optind;
static void	add2_logbuf(),
		dump_log();
static unsigned	set_logtype( );
char	rootdir[SMALLLEN],
	wn_tmpdir[sizeof(WN_TEMPDIR) + TINYLEN] = {WN_TEMPDIR},
	wn_empty[] = { "" },	/* safe place to park pointers */
	wnlogfile[SMALLLEN],
	errlogfile[SMALLLEN],
	pid_file[SMALLLEN],
	cfname[SMALLLEN],
	hostname[MAXHOSTNAMELEN],
	**mtypelist = NULL,
	**suflist = NULL;


int	admin_mode = FALSE,
	debug_log = FALSE,
	nofork = FALSE;


unsigned	serv_perm = 0,
		interface_num,
		default_logtype = WN_COMMON_LOG,
		acache_id,
		cache_id;

uid_t	user_id = USERID;
gid_t	group_id = GROUPID;


#ifndef NO_FLOCK
static void	locklog(),
		unlocklog();
#endif

static void	start_log(),
		restart();

static void	log_logfile();

static FILE	*logfp = NULL,
		*errlogfp = NULL;

static void	log_syslog ();

static void	append_gmtoff();





void
wn_init( argc, argv)
int	argc;
char	*argv[];
{
	char	*cp;
	int	c,
		errflg = 0;
		port = DEFAULT_PORT;
	mystrncpy( rootdir, ROOT_DIR, SMALLLEN);
	default_logtype = set_logtype( );
	if ( strlen( SWN_PID_FILE) > 0)
		mystrncpy( pid_file, SWN_PID_FILE, SMALLLEN);
	else
		*pid_file = '\0';
	mystrncpy( hostname, WN_HOSTNAME, MAXHOSTNAMELEN);

	if ( FORBID_CGI)
		serv_perm |= WN_FORBID_EXEC;
	if ((c = getopt(argc, argv, "p")) != -1) 
	{
		port = atoi( optarg);
	}
	if(errflag)
	{
		PRINT1_ERROR( "Unknown option given to server");
		exit (2);
	}

	if ( *wnlogfile && !*errlogfile) {
		mystrncpy( errlogfile, wnlogfile, SMALLLEN);
	}

	if ( argv[optind] )
		mystrncpy( rootdir, argv[optind], SMALLLEN);
		openlog ("wn", LOG_PID | LOG_NDELAY, LOGFACILITY);
	if ( ! STANDALONE )
		open_wnlog( wnlogfile, errlogfile);

#if STANDALONE init_mime();
#endif
}

void
open_wnlog( logfile, errlog)
char	*logfile,
	*errlog;
{

#if (! FORBID_CGI)
	char		buf[TINYLEN];
	struct stat	stat_buf;
	uid_t		my_id;

	my_id = getuid();
	Snprintf1( buf, TINYLEN, "/wn_tmp%d", my_id);
	mystrncat( wn_tmpdir, buf, sizeof(WN_TEMPDIR) + TINYLEN);
	if ( lstat( wn_tmpdir, &stat_buf) == -1) {
		if ( (errno != ENOENT) || (mkdir( wn_tmpdir, 0711) != 0) ) {
			PRINT_ERROR( err_m[122], wn_tmpdir);
			exit( 2);
		}
	}
	else if ( (!S_ISDIR(stat_buf.st_mode))
				|| (stat_buf.st_uid != my_id)
				|| (chmod( wn_tmpdir, 0711) != 0) ) {
		PRINT_ERROR( err_m[122], wn_tmpdir);
		PRINT_ERROR( "errno=%d:", errno);
		PRINT_ERROR( " %s\n", strerror( errno));
		exit( 2);
	}
#endif

	if ( *logfile != '\0')
		if ( (logfp = fopen( logfile, "a")) == NULL ) {
			PRINT_ERROR( err_m[0], logfile);
			PRINT_ERROR( "errno=%d:", errno);
			PRINT_ERROR( " %s\n", strerror( errno));
			if ( errno == ENOENT || errno == EACCES )
				PRINT_ERROR( " %s\n", err_m[152]);
			exit( 2);
		}
	if ( *errlog != '\0') {
		if ( streq( errlog, logfile))
			errlogfp = logfp;

		else if ( (errlogfp = fopen( errlog, "a")) == NULL ) {
			PRINT_ERROR( err_m[0], errlog);
			PRINT_ERROR( "errno=%d:", errno);
			PRINT_ERROR( " %s\n", strerror( errno));
			if ( errno == ENOENT || errno == EACCES )
				PRINT_ERROR( " %s\n", err_m[152]);
			exit( 2);
		}
	}
	if ( STANDALONE) {
		start_log( FALSE);
	}
}


static void
restart( )
{
#if  STANDALONE
#endif
	if ( *wnlogfile) {
		if ( logfp) 
			fclose( logfp);
		if ( (logfp = fopen( wnlogfile, "a")) == NULL ) {
			PRINT_ERROR( err_m[0], wnlogfile);
			PRINT_ERROR( "restart errno=%d:", errno);
			PRINT_ERROR( " %s\n", strerror( errno));
			if ( errno == ENOENT || errno == EACCES )
				PRINT_ERROR( " %s\n", err_m[152]);
			exit( 2);
		}
	}

	if ( *errlogfile) {
		if ( streq( errlogfile, wnlogfile))
			errlogfp = logfp;
		else {
			if ( errlogfp)
				fclose( errlogfp);
			if ( (errlogfp = fopen( errlogfile, "a")) == NULL ){
				PRINT_ERROR( err_m[0], errlogfile);
				PRINT_ERROR( "restart errno=%d:", errno);
				PRINT_ERROR( " %s\n", strerror( errno));
				if ( errno == ENOENT || errno == EACCES )
					PRINT_ERROR( " %s\n", err_m[152]);
				exit( 2);
			}
		}
	}
	start_log( TRUE);
	init_mime();
#endif
}


void
logerr(  msg, msg2)
char	*msg,
	*msg2;
{

#ifndef NO_FLOCK
	struct flock	lck;
#endif

	time_t	clock;
	struct tm *ltm;

	char	status[TINYLEN],
		date[TINYLEN],
		xmsg2[SMALLLEN + 2];

	unsigned llogtype;

	if ( (!dir_p) || !dir_p->logtype)
		llogtype = default_logtype;
	else 
		llogtype = dir_p->logtype;

	xmsg2[0] = '\0';
	if ( msg2 != NULL) {
		mystrncat2( xmsg2, msg2, SMALLLEN);
	}

	if ( llogtype & WN_NO_LOG)
		return;

	get_remote_info( );

 	if ( llogtype & (WN_LOG_SYSLOG + WN_VERBOSE_SYSLOG)) {
		log_syslog( ERRLOG_PRIORITY, this_rp, msg, xmsg2);
		return;
	}

	time(&clock);
	ltm = (struct tm *) localtime(&clock);
	strftime( date, TINYLEN, "%d/%h/%Y:%T", ltm);
	append_gmtoff( date, ltm);

#ifndef NO_FLOCK
	locklog( &lck, fileno( errlogfp) );
#endif

	if ( outheadp && (*outheadp->status)) {
		strncpy( status, outheadp->status, 4);
		status[4] = '\0';
		/* don't use mystrncpy as it may call us */
	}
	else
		strcpy( status, "500");

	fseek( errlogfp, 0L, 2);
	fprintf( errlogfp, "%.400s - - [%.64s] \"%.300s\" %.64s -",
		 this_conp->remotehost,	date, this_rp->request, status);
	fprintf( errlogfp, " <(%d/%d) %.400s: %.400s>",
			this_conp->pid, this_conp->trans_cnt, msg, xmsg2);

	fprintf( errlogfp, "\n");
	(void) fflush( errlogfp);


#ifndef NO_FLOCK
	unlocklog( &lck, fileno( errlogfp) );   
#endif

}

void
daemon_logerr(  msg, msg2, error)
char	*msg,
	*msg2;
int	error;
{

	FILE		*errfp;
#ifndef NO_FLOCK
	struct flock	lck;
#endif
	time_t	clock;
	struct tm *ltm;

	char	buf[2*SMALLLEN],
		date[TINYLEN];

	if ( default_logtype & WN_NO_LOG)
		return;

	if ( strlen( msg2) > SMALLLEN )  /* don't log too much */
		msg2[SMALLLEN-1] = '\0';

 	if ( default_logtype & (WN_LOG_SYSLOG + WN_VERBOSE_SYSLOG)) {
		mystrncpy( buf, 
		"none - - [] \"none\" 500 0  <%.100s: %.100s : %.100s> ");

	 	if ( default_logtype & WN_VERBOSE_SYSLOG) {
			mystrncat2( buf, "<> <> <>\n", SMALLLEN);
		}
		else
			mystrncat( buf, "\n", SMALLLEN);

		syslog( ERRLOG_PRIORITY, buf, msg, msg2, strerror( error));
		return;
	}

	if ( errlogfp)
		errfp = errlogfp;
	else
		errfp = stderr;

	time(&clock);
	ltm = (struct tm *) localtime(&clock);
	strftime( date, TINYLEN, "%d/%h/%Y:%T", ltm);
	append_gmtoff( date, ltm);

#ifndef NO_FLOCK
	if ( errlogfp)
		locklog( &lck, fileno( errlogfp) );
#endif

	fseek( errfp, 0L, 2);   
	fprintf(errfp, "none - - [%.64s] \"none\" 500 0",  date);
	if ( default_logtype & WN_VERBOSE_LOG ) {
		fprintf( errfp, " <%.200s %.200s: %s> <> <> <>",
		msg, msg2, strerror(error));
	}
	fprintf( errfp, "\n");
	(void) fflush( errfp);

#ifndef NO_FLOCK
	if ( errlogfp)
		unlocklog( &lck, fileno( errlogfp) );   
#endif
}


void
writelog( ip, msg, msg2)
Request	*ip;
char	*msg,
	*msg2;
{
	char	xmsg2[SMALLLEN];
	unsigned llogtype;

	if ( (!dir_p) || !dir_p->logtype)
		llogtype = default_logtype;
	else 
		llogtype = dir_p->logtype;

 	if ( llogtype & WN_NO_LOG)
		return;

	xmsg2[0] = '\0';
	if ( msg2 != NULL)
		mystrncat2( xmsg2, msg2, SMALLLEN);

 	if ( llogtype & (WN_LOG_SYSLOG + WN_VERBOSE_SYSLOG))
		log_syslog( LOG_PRIORITY, ip, msg, xmsg2);
	else
		log_logfile( ip, msg, xmsg2);
}


static void
log_logfile( ip, msg, msg2)
Request	*ip;
char	*msg,
	*msg2;
{
	time_t	clock;
	struct tm *ltm;
	char	*authname,
		bytes[TINYLEN],
		status[TINYLEN],
		date[TINYLEN],
		lbuf[2*SMALLLEN];

	unsigned llogtype;
	if ( (!dir_p) || !dir_p->logtype)
		llogtype = default_logtype;
	else 
		llogtype = dir_p->logtype;
	if ( outheadp && (*outheadp->status)) {
		char *cp;
		status[0] = '\0';
		mystrncat2( status, outheadp->status, TINYLEN);
		cp = status;
		while ( *cp && !isspace(*cp))
			cp++;
		*cp = '\0';
	}
	else
		mystrncpy( status, "200", TINYLEN);

	Snprintf1( bytes, TINYLEN, "%ld", ip->logcount);

	time(&clock);
	ltm = (struct tm *) localtime(&clock);
	strftime( date, TINYLEN, "%d/%h/%Y:%T", ltm);
	append_gmtoff( date, ltm);

	authname = ( *this_rp->authuser ? this_rp->authuser : "-");
	fmt3( lbuf, 2*SMALLLEN, authname, " [", date);
	fmt3( lbuf, 2*SMALLLEN, lbuf, "] \"", ip->request);
	fmt3( lbuf, 2*SMALLLEN, lbuf, "\" ", status);
	fmt3( lbuf, 2*SMALLLEN, lbuf, " ", bytes);

	add2_logbuf( lbuf);

	if ( llogtype & WN_VERBOSE_LOG ) {
		if ( streq( status, "206")) {
			char *rangep;
			Snprintf5( lbuf, 2*SMALLLEN, 
				" <(%d/%d) %.100s (%.100s): %.200s>",
				this_conp->pid, this_conp->trans_cnt, 
				log_m[23], rangep, msg2);
		}
		else {
			Snprintf4( lbuf, 2*SMALLLEN, " <(%d/%d) %.200s: %.200s>",
			this_conp->pid, this_conp->trans_cnt, msg, msg2);
		}
		add2_logbuf( lbuf);
		fmt3( lbuf, 2*SMALLLEN, " <", inheadp->referrer, ">");
		add2_logbuf( lbuf);
	}
	}
	mystrncat2( this_conp->logbuf, "\n", LOGBUFLEN);
}

static void
add2_logbuf( xlbuf)
char	*xlbuf;
{

	if ( strlen( xlbuf) + strlen( this_conp->logbuf) >= LOGBUFLEN) {
		dump_log( );
	}
	mystrncat2( this_conp->logbuf, xlbuf, LOGBUFLEN);
	xlbuf[0] = '\0';
}

#ifndef NO_FLOCK
static void
locklog(lck, fd)
struct flock	*lck;
int	fd;
{
	lck->l_type = F_WRLCK;
	lck->l_whence = SEEK_SET;
	lck->l_start = 0L;
	lck->l_len = 0L;
	fcntl(fd, F_SETLKW, lck);
}

static void
unlocklog( lck, fd)
struct flock	*lck;
int	fd;
{
	lck->l_type = F_UNLCK;
	fcntl( fd, F_SETLKW, lck);
}
#endif


static void
log_syslog( priority, ip, msg, msg2)
int	priority;
Request	*ip;
char	*msg,
	*msg2;
{
	char	*authname,
		bytes[TINYLEN],
		status[TINYLEN];
	unsigned llogtype;
	if ( (!dir_p) || !dir_p->logtype)
		llogtype = default_logtype;
	else 
		llogtype = dir_p->logtype;

	if ( *outheadp->status) {
		strncpy( status, outheadp->status, 4);
		/* don't use mystrncpy as it may call us */
		status[3] = '\0';
	}
	else {
		strcpy( status, "200");
	}

	Snprintf1( bytes, TINYLEN, "%ld", ip->logcount);
	authname = ( *ip->authuser ? ip->authuser : "-");
	get_remote_info();
	if ( llogtype & WN_VERBOSE_SYSLOG ) {
		char 	xlbuf[SMALLLEN];
		syslog( priority,
		"%.100s %.32s %.32s \"%.100s\" %.10s %.10s  <%.100s: %.100s> <%.100s> <%.100s> <%.100s>%s\n", 
		this_conp->remotehost, rfc931p, authname, ip->request, status, 
		bytes, msg, msg2, inheadp->ua, inheadp->referrer,
		inheadp->cookie, xlbuf);
	}
	else {
		syslog( priority,
		"%.100s %.32s %.32s \"%.100s\" %.10s %.10s  <%.100s: %.100s>\n", 
		this_conp->remotehost, rfc931p, authname, ip->request, status, 
		bytes, msg, msg2);
	}
}



static void
append_gmtoff( date, ltm)
char	*date;
struct tm *ltm;
{
	register char	*cp;
	long		tz;
	char 		sign;

	tz = - timezone;
	if( ltm->tm_isdst)
		tz += 3600;
	sign = ( tz > 0 ? '+' : '-');
	tz = ( tz > 0 ? tz :  -tz);
	cp = date;
	while ( *cp)
		cp++;
	
	Snprintf3( cp, 7, " %c%02ld%02ld", sign, tz/3600, tz % 3600);
}

static void
start_log( restart)
int	restart;
{
#if STANDALONE
	time_t	clock;
	struct tm *ltm;
	char	*cp,
		startdate[TINYLEN];
	if ( default_logtype & (WN_NO_LOG+WN_LOG_SYSLOG+WN_VERBOSE_SYSLOG))
		return;
	time(&clock);
	ltm = (struct tm *) localtime(&clock);
	strftime( startdate, TINYLEN, "%d/%h/%Y:%T", ltm);
	cp = ( restart ? "Restarting" : "Starting");
	fprintf(logfp, "\n%.64s: %s %.64s at port %d with pid %d\n", 
			startdate, cp, VERSION, port, getpid() );
	(void) fflush(logfp);
#endif
}


void
wn_abort( )
{

	this_conp->keepalive = FALSE;
	this_rp->type = RTYPE_FINISHED;
	this_rp->status |= (WN_ABORTED + WN_ERROR);
	
}

void
wn_exit( status)
int	status;
{
	int	n;
	char	buf[MIDLEN];
	if ( this_conp && (this_conp->chunk_status & WN_USE_CHUNK))
		end_chunking();
	flush_outbuf();
	if ( inheadp->tmpfile_name && *inheadp->tmpfile_name) {
		unlink( inheadp->tmpfile_name);
		*inheadp->tmpfile_name = '\0';
	}
	shutdown( fileno( stdin), 1);
	dump_log();
	if ( status != 0) {
		/* error condition */
		/* Give ourselves at most 3 seconds to shutdown */
		while ( (n = WN_read( (fileno( stdin)), buf, MIDLEN)) > 0 ) {
			; /* Collect stuff from client and throw it away */
		}
	}

	close( fileno( stdin));
	exit( status);
}


static void
dump_log()
{
	register char	*cp,
			*cp2;
	static int	looping = FALSE;
#ifndef NO_FLOCK
	struct flock	lck;
#endif

	if ( (! this_conp) || *(this_conp->logbuf) == '\0') {
		return;
	}
	get_remote_info( );
#ifndef NO_FLOCK
	locklog( &lck, fileno( logfp) );
#endif

	fseek( logfp, 0L, 2);
	cp = this_conp->logbuf;
	while ( (cp2 = strchr( cp, '\n')) ) {
		char save_ch;
		save_ch = *++cp2;
		*cp2 = '\0';
		fprintf( logfp, "%.255s %.100s %.1000s", 
			 this_conp->remotehost, rfc931p, cp);
		cp = cp2;
		*cp = save_ch;
		looping = FALSE;
	}
	if ( *cp && cp > this_conp->logbuf ) 
		mystrncpy( this_conp->logbuf, cp, LOGBUFLEN);
	if ( !*cp || looping ) {
		*(this_conp->logbuf) = '\0';
		looping = FALSE;
	}
	else if ( cp == this_conp->logbuf  )
		looping = TRUE;
	(void) fflush( logfp);
#ifndef NO_FLOCK
	unlocklog( &lck, fileno( logfp) );   
#endif
}
void
get_mtype( suffix )
char	*suffix;
{
	int	i,
		unknown_mime = TRUE;
	for ( i = 0; mimelist[i][0] != NULL; i++) {
		if ( streq( mimelist[i][0], suffix)) {
			mystrncpy( this_rp->contype, mimelist[i][1], SMALLLEN);
			this_rp->content_type = this_rp->contype;
			unknown_mime = FALSE;
			break;
		}
	}
	if ( STANDALONE && unknown_mime && (suflist != NULL)) {
		for ( i = 0; suflist[i] != NULL; i++) {
			if ( streq( suflist[i], suffix)) {
				mystrncpy( this_rp->contype, 
						mtypelist[i], SMALLLEN);
				this_rp->content_type = this_rp->contype;
				break;
			}
		}
	}
}

#if STANDALONE
int
needsuffix( suf)
char	*suf;
{
	int	i;
	for ( i = 0; mimelist[i][0] != NULL; i++) {
		if ( streq( suf, mimelist[i][0]))
			return FALSE;
	}
	return TRUE;
}
#endif  /* STANDALONE */

static unsigned
set_logtype( )
{
	unsigned logtype = 0;

	if ( USE_NO_LOG)
		logtype = WN_NO_LOG;

	else if ( USE_SYSLOGD )
		logtype = WN_LOG_SYSLOG;

	else if ( USE_FILE_LOG)
		logtype = WN_COMMON_LOG;

	if ( logtype == 0)
		logtype = WN_NO_LOG;

	if ( logtype == WN_NO_LOG)
		logtype |= NO_DNS_LOG;

	mystrncpy (wnlogfile, WN_LOGFILE, SMALLLEN);
	mystrncpy (errlogfile, WN_ERRLOGFILE, SMALLLEN);

	return (logtype);
}